==> 1.6.92.tar.gz: lib/src/mixin_history.dart <==
part of '../flutterx_live_data.dart';

/// A [LiveData] mixin which keeps track of each modification to it's value and allows to undo or redo changes
///
/// Note: in order to use this LiveData, the value must be serialized and deserialized.
/// This is done by [encode] and [decode] functions
///
/// To know you can undo/redo the history you can query [canUndo] and [canRedo] value
/// Once you are sure you can execute a certain action you can call [historyUndo] and [historyRedo] methods to
/// undo/redo a change made to this LiveData, or you can call anytime [historyClear] to completely clear the history
///
/// You also can set [historyCapacity] field which determines the max size of the history (by default is 20)
/// If you set a length that is less than the current history queue the first elements of the history will be
/// removed until the history length matches the desired value
mixin HistoryMixin<T> on LiveData<T> {
  /// Queue of all saved states of this LiveData
  @protected
  final Queue<String> queue = Queue();
  int _capacity = 20;
  int _index = -1;
  bool _restoring = false;

  /// Tells if you can safely call [historyUndo].
  final LiveData<bool> canUndo = MutableLiveData(initialValue: false);

  /// Tells if you can safely call [historyRedo].
  final LiveData<bool> canRedo = MutableLiveData(initialValue: false);

  /// Get max length of history queue
  int get historyCapacity => _capacity;

  /// Set max length of history queue and automatically resize history to match desired length
  set historyCapacity(int value) {
    if (value < 1) throw ArgumentError.value(value, 'historyCapacity', 'Cannot be less than 1');
    _capacity = value;
    if (queue.length >= value) {
      queue.removeFirst();
      while (queue.length >= value) queue.removeFirst();
      if (historyIndex >= queue.length)
        historyIndex = queue.length - 1;
      else
        updateValues();
    }
  }

  /// Get current index of the history, if this index is -1 it means [queue] is empty
  int get historyIndex => _index;

  /// Set current history index, the value must be within inclusive range of -1 and [historyLength]-1
  set historyIndex(int value) {
    if (value == _index) return;
    if (value < -1 || value >= queue.length) throw IndexError(value, queue, 'historyIndex');
    _index = value;
    if (value >= 0) restoreState();
    updateValues();
  }

  /// Get current history queue length
  int get historyLength => queue.length;

  /// Convert given value to string defaults to [jsonEncode]
  @protected
  String Function(T data)? get encode => null;

  /// Convert given string to value defaults to [jsonDecode]
  @protected
  T Function(String source)? get decode => null;

  @override
  void invalidate() {
    if (!_restoring) {
      saveState();
      updateValues();
    }
    super.invalidate();
  }

  /// Reset current LiveData value to previous history value
  void historyUndo() {
    if (!canUndo._data) throw StateError('Cannot undo');
    historyIndex--;
  }

  /// Reset current LiveData value to next history value
  void historyRedo() {
    if (!canRedo._data) throw StateError('Cannot redo');
    historyIndex++;
  }

  /// Undo and delete all entries after current one
  void historyDelete() {
    if (queue.isEmpty) throw StateError('Cannot delete');
    while (historyIndex <= queue.length - 1) queue.removeLast();
    historyIndex--;
  }

  /// Replace current entry with new data
  void historyReplace(T data) {
    if (queue.isEmpty) throw StateError('Cannot replace');
    historyDelete();
    setData(data);
  }

  /// Clear history queue. This will not alter current LiveData value
  void historyClear() {
    queue.clear();
    historyIndex = -1;
  }

  /// Save current state into queue
  @protected
  @mustCallSuper
  void saveState() {
    final state = (encode ?? jsonEncode)(_data);
    final index = historyIndex;
    if (index > -1 && state == queue.elementAt(index)) return;
    while (index < queue.length - 1) queue.removeLast();
    queue.add(state);
    if (index + 1 >= _capacity)
      queue.removeFirst();
    else
      historyIndex++;
  }

  /// Restore previously saved state
  @protected
  @mustCallSuper
  void restoreState() {
    final state = (decode ?? jsonDecode)(queue.elementAt(historyIndex));
    try {
      _restoring = true;
      setData(state);
    } finally {
      _restoring = false;
    }
  }

  /// Update [canUndo] and [canRedo] fields
  @protected
  void updateValues() {
    final index = historyIndex;
    canUndo.setData(index > 0);
    canRedo.setData(index < queue.length - 1);
  }
}

==> 1.7.2.tar.gz: lib/src/mixin_history.dart <==
part of '../flutterx_live_data.dart';

/// A [LiveData] mixin which keeps track of each modification to it's value and allows to undo or redo changes
///
/// Note: in order to use this LiveData, the value must be serialized and deserialized.
/// This is done by [encode] and [decode] functions
///
/// To know you can undo/redo the history you can query [canUndo] and [canRedo] value
/// Once you are sure you can execute a certain action you can call [historyUndo] and [historyRedo] methods to
/// undo/redo a change made to this LiveData, or you can call anytime [historyClear] to completely clear the history
///
/// You also can set [historyCapacity] field which determines the max size of the history (by default is 20)
/// If you set a length that is less than the current history queue the first elements of the history will be
/// removed until the history length matches the desired value
mixin HistoryMixin<T> on LiveData<T> {
  /// Queue of all saved states of this LiveData
  @protected
  final Queue<String> queue = Queue();
  int _capacity = 20;
  int _index = -1;
  bool _restoring = false;

  /// Tells if you can safely call [historyUndo].
  final LiveData<bool> canUndo = MutableLiveData(initialValue: false);

  /// Tells if you can safely call [historyRedo].
  final LiveData<bool> canRedo = MutableLiveData(initialValue: false);

  /// Get max length of history queue
  int get historyCapacity => _capacity;

  /// Set max length of history queue and automatically resize history to match desired length
  set historyCapacity(int value) {
    if (value < 1) throw ArgumentError.value(value, 'historyCapacity', 'Cannot be less than 1');
    _capacity = value;
    if (queue.length >= value) {
      queue.removeFirst();
      while (queue.length >= value) queue.removeFirst();
      if (_index >= queue.length)
        historyIndex = queue.length - 1;
      else
        updateValues();
    }
  }

  /// Get current index of the history, if this index is -1 it means [queue] is empty
  int get historyIndex => _index;

  /// Set current history index, the value must be within inclusive range of -1 and [historyLength]-1
  set historyIndex(int value) {
    if (value == _index) return;
    if (value < -1 || value >= queue.length) throw IndexError(value, queue, 'historyIndex');
    _index = value;
    if (value >= 0) restoreState();
    updateValues();
  }

  /// Get current history queue length
  int get historyLength => queue.length;

  /// Convert given value to string defaults to [jsonEncode]
  @protected
  String Function(T data)? get encode => null;

  /// Convert given string to value defaults to [jsonDecode]
  @protected
  T Function(String source)? get decode => null;

  @override
  void invalidate() {
    if (!_restoring) {
      saveState();
      updateValues();
    }
    super.invalidate();
  }

  /// Reset current LiveData value to previous history value
  void historyUndo() {
    if (!canUndo._data) throw StateError('Cannot undo');
    historyIndex--;
  }

  /// Reset current LiveData value to next history value
  void historyRedo() {
    if (!canRedo._data) throw StateError('Cannot redo');
    historyIndex++;
  }

  /// Undo and delete all entries after current one
  void historyDelete() {
    if (queue.isEmpty) throw StateError('Cannot delete');
    while (queue.length - 1 >= _index) queue.removeLast();
    historyIndex--;
  }

  /// Replace current entry with new data
  void historyReplace(T data, {int? index}) {
    if (queue.isEmpty) throw StateError('Cannot replace');
    index ??= _index;
    if (index > _index) throw IndexError(index, queue, 'index');
    while (queue.length - 1 >= index) queue.removeLast();
    _index = index - 1;
    _data = data;
    invalidate();
  }

  /// Clear history queue. This will not alter current LiveData value
  void historyClear() {
    queue.clear();
    historyIndex = -1;
  }

  /// Save current state into queue
  @protected
  @mustCallSuper
  void saveState() {
    final state = (encode ?? jsonEncode)(_data);
    final index = _index;
    if (index > -1 && state == queue.elementAt(index)) return;
    while (index < queue.length - 1) queue.removeLast();
    queue.add(state);
    if (index + 1 >= _capacity)
      queue.removeFirst();
    else
      historyIndex++;
  }

  /// Restore previously saved state
  @protected
  @mustCallSuper
  void restoreState() {
    final state = (decode ?? jsonDecode)(queue.elementAt(_index));
    try {
      _restoring = true;
      _data = state;
      invalidate();
    } finally {
      _restoring = false;
    }
  }

  /// Update [canUndo] and [canRedo] fields
  @protected
  void updateValues() {
    final index = _index;
    canUndo.setData(index > 0);
    canRedo.setData(index < queue.length - 1);
  }
}
